#!/usr/bin/perl
use strict;

### 0dbfs.pl
#   Bringing a nuke to the loudness war
#   Greg Kennedy 2012
#
#  Usage: ./0dbfs.pl in.wav out.wav

# Minimum threshhold for sample size, to help cut transients
#  Windows smaller than $thresh will be zeroed
my $thresh = 3;
#  Peaks smaller than $minpeak will also zero
my $minpeak = 0.05;

# function to read and verify we read correct bytes
sub r_fp_n
{
  my $FP = shift;
  my $bytes = shift;

  my $buf;

  my $bytesread = read ($FP, $buf, $bytes);
  if ($bytesread != $bytes)
  {
    die "Short read: expected $bytes, read $bytesread\n";
  }
  return $buf;
}

my $FP;
open($FP, $ARGV[0]) or die "can't open $ARGV[0]: $!\n";
binmode($FP);

# read RIFF header
my $temp = r_fp_n($FP,4);
if ($temp ne 'RIFF') {
  die "Does not appear to be a RIFF file: header $temp\n";
}

my $chunksize = unpack('V',r_fp_n($FP,4));
print "Info: chunksize = " . $chunksize . "\n";

$temp = r_fp_n($FP,4);
if ($temp ne 'WAVE')
{
  die "Does not appear to be a RIFF-WAVE file: format $temp\n";
}

$temp = r_fp_n($FP,4);
if ($temp ne 'fmt ')
{
  die "Expected: subchunk 'fmt ', got: subchunk $temp\n";
}

$temp = unpack('V',r_fp_n($FP,4));
if ($temp != 16)
{
  die "Unhappy with fmt-subchunk size $temp, bailing out\n";
}

$temp = unpack('v',r_fp_n($FP,2));
if ($temp != 1)
{
  die "This seems to be a non-PCM file (e.g. ADPCM, etc), please convert to God's own format\n";
}

my $channels = unpack('v',r_fp_n($FP,2));
print "Info: $channels channels\n";

my $samplerate = unpack('V',r_fp_n($FP,4));
print "Info: samplerate $samplerate\n";

my $byterate = unpack('V',r_fp_n($FP,4));
print "Info: byterate: $byterate\n";
my $blockalign = unpack('v',r_fp_n($FP,2));
print "Info: blockalign: $blockalign\n";

my $bits = unpack('v',r_fp_n($FP,2));
print "Info: $bits bit waveform\n";

# data time
my $temp = r_fp_n($FP,4);
if ($temp ne 'data')
{
  die "Expected: subchunk 'data', got: '$temp'\n";
}

my $chsize = unpack('V',r_fp_n($FP,4));
print "Info: data chunk size: $chsize\n";

# Slurp entire wavefile.
my @waveform = ();

for (my $i = 0; $i < $chsize; $i += $blockalign)
{
  for (my $j = 0; $j < $channels; $j ++)
  {
    if ($bits == 8)
    {
      my $val = (ord(r_fp_n($FP,1)) - 128);
      if ($val < 0) { $val /= 128.0; } else { $val /= 127.0; }
      push(@{$waveform[$j]},$val);
    } elsif ($bits == 16)
    {
      my $val = unpack('s<',r_fp_n($FP,2));
      if ($val < 0) { $val /= 32768.0; } else { $val /= 32767.0; }
      push(@{$waveform[$j]},$val);
    } elsif ($bits == 24)
    {
      # read as unsigned
      my $val = unpack('V',r_fp_n($FP,3) . chr(0));
      # convert to signed
      if ($val >= (2 ** 23)) { $val = ($val - (2 ** 24)); }
      # convert to float
      if ($val < 0) { $val /= (2 ** 23); } else { $val /= ((2 ** 23) - 1); }
      push(@{$waveform[$j]},$val);
    }
  } 
}

close(FP);

##### search for peaks
print "Info: file read complete.\n";

for (my $j = 0; $j < $channels; $j ++)
{
  my $seg_start = 0;
  my $peak = 0;
  my $sign = 0;

  my $i;

  print "Info: processing on " . scalar @{$waveform[$j]} . " samples, channel $j\n";

  for ($i = 0; $i < scalar @{$waveform[$j]}; $i ++)
  {
    my $val = $waveform[$j][$i];
    if ((($val > 0) != $sign) || $i == scalar @{$waveform[$j]} - 1)
    {
      if (($i - $seg_start < $thresh) || $peak < $minpeak)
      {
        print "CROSS: sample window " . ($i - $seg_start) . " is below threshold.\n";
        for (my $k = $seg_start; $k < $i; $k ++)
        {
          $waveform[$j][$k] = 0;
        }
      } else {
        # zero-crossing
        my $mult = 0;
        if ($peak != 0)
        {
          $mult = 1.0 / $peak;
        }
print "CROSS: sample window " . ($i - $seg_start) . ", peak value " . $peak . ", scaling by " . $mult . "\n";
        for (my $k = $seg_start; $k < $i; $k ++)
        {
          $waveform[$j][$k] *= $mult;
        }
      }
      $seg_start = $i;
      $peak = abs($val);
      $sign = ($val > 0);
    } else {
      if (abs($val) > $peak)
      {
          $peak = abs($val);
      }
    }
  }
}

##### dump 'radio-ready' waveform
my $FP2;

open($FP2,">$ARGV[1]") or die "can't open $ARGV[1]:$!\n";
binmode($FP2);

print $FP2 'RIFF' . pack('V',$chunksize) . 'WAVE';
print $FP2 'fmt ' . pack('V',16) . pack('v',1) . pack('v',$channels) . pack('V',$samplerate) . pack('V',$byterate) . pack('v',$blockalign) . pack('v', $bits);
print $FP2 'data' . pack('V',$chsize);

for (my $i = 0; $i < scalar @{$waveform[0]}; $i ++)
{
  for (my $j = 0; $j < $channels; $j ++)
  {
    my $val = $waveform[$j][$i];
    if ($bits == 8)
    {
      if ($val < 0) { $val *= 128.0; } else { $val *= 127.0; }
      $val += 128;
      print $FP2 chr(int($val));
    } elsif ($bits == 16)
    {
      if ($val < 0) { $val *= 32768.0; } else { $val *= 32767.0; }
      print $FP2 pack('s',int($val));
    } elsif ($bits == 24)
    {
      # float -> int
      if ($val < 0) { $val *= (2 ** 23); } else { $val *= ((2 ** 23) - 1); }
      # make signed
      if ($val < 0) { $val += (2 ** 24); }
      print $FP2 substr(pack('V',int($val)),0,3);
    }
  }
}

close($FP2);
